import random
import re
import uuid

from collections import namedtuple
from string import ascii_letters, ascii_uppercase
from typing import List, Optional

try:
    from typing import Literal  # Only available on 3.8+ directly
except ImportError:
    from typing_extensions import Literal

FQTN = namedtuple("FQTN", ["database", "schema", "table"], defaults=(None, None, None))


def is_fqtn(fqtn: str, raise_if_false: bool = False) -> bool:
    """
    Determine if a string is a fully qualified table name

    Allowed formats:
    DATABASE.SCHEMA.TABLE
    "DATABASE"."SCHEMA"."TABLE"
    `project:dataset.table`
    `project-001`.`dataset`.`table`

    Returns a boolean, or raises an exception if directed
    """
    if fqtn and re.match(r'["`\-\w]+[\.:]["`\w]+\.["`\w]+', fqtn):
        return True
    if raise_if_false:
        raise ValueError(f"{fqtn} is not a well-formed fqtn")
    return False


def make_fqtn(
    table: str, database: Optional[str] = None, schema: Optional[str] = None, org_defaults: dict = None
) -> str:
    if is_fqtn(table):
        return table
    if not org_defaults and (not database or not schema):
        raise ValueError("Need to pass database and schema to make a FQTN")
    database = database or org_defaults.get("database")
    schema = schema or org_defaults.get("schema")
    return f"{database}.{schema}.{table}"


def split_fqtn(fqtn: str, org_defaults: dict = None) -> FQTN:
    """
    Accepts a possible fully qualified table string and returns its component parts
    """
    database = org_defaults.get("database") if org_defaults else None
    schema = org_defaults.get("schema") if org_defaults else None
    table = fqtn
    if fqtn.count(".") == 2:
        database, schema, table = fqtn.split(".")
    if fqtn.count(".") == 1:
        schema, table = fqtn.split(".")
    return FQTN(database, schema, table)


def quote_fqtn(
    fqtn: str,
    dw_type: str = Literal['bigquery', 'snowflake'],
) -> str:
    """
    Handle quoting of fqtns
    """
    quote_char = '`' if dw_type == 'bigquery' else '"'
    if is_fqtn(fqtn, raise_if_false=True):
        database, schema, table = split_fqtn(fqtn)
        database = f"{quote_char}{database}{quote_char}" if needs_quoting(database) else database
        schema = f"{quote_char}{schema}{quote_char}" if needs_quoting(schema) else schema
        table = f"{quote_char}{table}{quote_char}" if needs_quoting(table) else table
        return f"{database}.{schema}.{table}"


def needs_quoting(token: str):
    """
    Does this token contain characters that need quoting in the DW

    Returns a boolean
    """
    # NOTE: we may want to use ascii_uppercase here when we can trust input data to be correctly casesd
    if any([char for char in token if char not in (ascii_letters + "_")]):
        return True
    if token in (SNOWFLAKE_RESTRICTED_KEYWORDS):
        return True
    return False


def gen_operation_table_name(
    op_num: int,
    transform_name: str,
) -> str:
    """
    Generate and return the table or view name to set as the
    output of an operation.

    We try to create as intelligent as a name possible with the information
    we have about the offline dataset at the time of transformation.

    GUID is added to ensure uniqueness of table/view names

    Args:
        op_num: Number of this operation in it's operation op set
        transform_name: Name of the transform applied

    Returns:
        Table or View name to set as output for a operation
    """
    short_guid = str(uuid.uuid4()).replace('-', '')[:10]
    return f"RASGO_SDK__OP{op_num}__{transform_name}_transform__{short_guid}".upper()


def random_alias() -> str:
    """
    Returns a random 16 char string
    """
    return ''.join(random.choice(ascii_uppercase) for x in range(16))


def cleanse_sql_data_type(dtype: str) -> str:
    """
    Converts a string to Snowflake compliant data type
    """
    if dtype.lower() in ["object", "text", "variant"]:
        return "string"
    else:
        return dtype.lower()


def cleanse_sql_list(list_in: List[str]) -> List[str]:
    """
    Converts a list of strings to Snowflake compliant names
    """
    return [cleanse_sql_name(n) for n in list_in]


def cleanse_sql_name(name: str) -> str:
    """
    Converts a string to a snowflake compliant value
    """
    name = name.replace(" ", "_").replace("-", "_").replace('"', '').replace(".", "_").upper()
    if name[0] not in (ascii_letters + "_"):
        return f"_{name}"
    return name


def cleanse_dbt_name(name: str) -> str:
    """
    Converts a string to a dbt compliant value
    """
    return name.replace(" ", "_").replace("-", "_").replace('"', '').replace(".", "_").lower()


def is_restricted_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of dangerous keywords
    """
    if any(word in sql.upper() for word in SNOWFLAKE_RESTRICTED_KEYWORDS):
        return True
    return False


def is_scary_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of injection keywords
    """
    if any(word in sql.upper() for word in SQL_INJECTION_KEYWORDS):
        return True
    return False


def is_valid_view_sql(sql: str) -> bool:
    """
    Checks a SQL string for presence of structural keywords
    """
    mandatory_words = ["SELECT", "FROM"]
    if not all(word in sql.upper() for word in mandatory_words):
        return False
    return True


SQL_RESTRICTED_CHARACTERS = [' ', '-', ';']

SQL_INJECTION_KEYWORDS = [
    'DELETE',
    'TRUNCATE',
    'DROP',
    'ALTER',
    'UPDATE',
    'INSERT',
    'MERGE',
]

# fmt: off
SNOWFLAKE_RESTRICTED_KEYWORDS = [
    'ACCOUNT','ALL','ALTER','AND','ANY','AS','BETWEEN','BY',
    'CASE','CAST','CHECK','COLUMN','CONNECT','CONNECTION','CONSTRAINT',
    'CREATE','CROSS','CURRENT','CURRENT_DATE','CURRENT_TIME','CURRENT_TIMESTAMP',
    'CURRENT_USER','DATABASE','DELETE','DISTINCT','DROP','ELSE','EXISTS','FALSE',
    'FOLLOWING','FOR','FROM','FULL','GRANT','GROUP','GSCLUSTER','HAVING','ILIKE',
    'IN','INCREMENT','INNER','INSERT','INTERSECT','INTO','IS','ISSUE','JOIN','LATERAL',
    'LEFT','LIKE','LOCALTIME','LOCALTIMESTAMP','MINUS','NATURAL','NOT','NULL','OF','ON',
    'OR','ORDER','ORGANIZATION','QUALIFY','REGEXP','REVOKE','RIGHT','RLIKE','ROW','ROWS',
    'SAMPLE','SCHEMA','SELECT','SET','SOME','START','TABLE','TABLESAMPLE','THEN','TO','TRIGGER',
    'TRUE','TRY_CAST','UNION','UNIQUE','UPDATE','USING','VALUES','VIEW','WHEN','WHENEVER','WHERE','WITH'
]
# fmt: on



import time
from typing import Optional

from pyrasgo import errors, schemas


def poll_operation_set_async_status(
    task_request: schemas.OperationSetAsyncTask,
    delay_start=1,
    delay_max=1.5,
    backoff_multiplier=1.5,
    verbose=False,
) -> int:
    """
    Given some task request, this function will poll the API until the task hits a terminal status.

    If operation set created successfully with no errors, return operation set id
    """
    from pyrasgo.api import Get

    get = Get()

    operations_expected = len(task_request.request.operations)
    operations_created = 0
    delay = delay_start
    while True:
        time.sleep(delay)
        task_request = get._operation_set_async_status(task_request.id)

        # Get the status of the last event
        last_event_status = task_request.events[-1].event_type

        # Let the user know if this started
        if last_event_status == 'STARTED':
            if verbose:
                print(f'Operation set task {task_request.id} has started')

        # Is this progress? Keep track.
        if last_event_status == 'PROGRESSED':
            operations_created = task_request.events[-1].message['complete']
            if verbose:
                print(f'Operation set creation: {operations_created} of {operations_expected} operations complete')

        # Is this a failure? Throw an exception
        if last_event_status == 'FAILED':
            raise errors.APIError(
                f'Operation set creation failure: {task_request.events[-1].message}; '
                f'Progress: {operations_created} / {operations_expected}'
            )

        # Is this a success? Get out of this loop
        if last_event_status == 'FINISHED':
            operation_set_id = task_request.events[-1].message['operation_set_id']
            return operation_set_id

        # Still waiting? Grab a Snickers
        if delay < delay_max:
            delay = min(delay * backoff_multiplier, delay_max)
        if verbose:
            print('Poll waiting...')


def poll_dataset_publish(
    connection_obj,
    max_poll_attempts: int,
    status_tracking_obj: schemas.StatusTracking,
    poll_retry_rate: int,
    timeout: Optional[int] = None,
) -> schemas.Dataset:
    """
    When publishing a dataset async, this function tracks StatusTracking
    objects and returns a dataset, if one is successfully created

    Args:
        connection_obj: API object for calls
        max_poll_attempts: The maxiumum number of poll attempts
        status_tracking_obj: A StatusTracking schema to use for the first iteration of calls
        timeout: Max timeout window after which to fail the dataset publishing
        poll_retry_rate: Number of seconds between which to make another status tracking check
    """
    for i in range(1, max_poll_attempts):
        status_tracking_obj = schemas.StatusTracking(
            **connection_obj._get(f"/status-tracking/{status_tracking_obj.tracking_uuid}", api_version=2).json()
        )
        if status_tracking_obj.status == "completed":
            # TODO:
            # This try-except is here because in the API, some status tracking
            # messages return the dataset ID, and some return the dataset
            # object. We've got parsing in the SDK for now, but it sure would
            # be nice to just make this uniform in the API
            try:
                ds_id = int(status_tracking_obj.message)
            except ValueError:
                if status_tracking_obj.message.count("(") != 1:
                    raise errors.APIError(
                        "Status tracking object malformed - expecting string representation of Dataset Class"
                    )
                ds_id = int(
                    status_tracking_obj.message[
                        status_tracking_obj.message.find("(") + 1 : status_tracking_obj.message.find(")")
                    ]
                )
            return schemas.Dataset(**connection_obj._get(f"/datasets/{ds_id}", api_version=2).json())
        if status_tracking_obj.status == "failed":
            raise errors.APIError(f"Could not publish dataset: {status_tracking_obj.message}")
        if timeout and (poll_retry_rate * i) > timeout:
            raise errors.APITimeoutWarning("Timeout reached waiting for dataset creation.")
        time.sleep(poll_retry_rate)
    raise errors.APIError(
        f"Never received confirmation dataset was published "
        f"after max wait of {(max_poll_attempts * poll_retry_rate) // 60} minutes."
    )



"""
API Create Functions
"""
from typing import Any, Dict, List, Optional, Union
from urllib import parse

try:
    from typing import Literal  # Only available on 3.8+ directly
except ImportError:
    from typing_extensions import Literal

from pyrasgo.config import MAX_POLL_ATTEMPTS, POLL_RETRY_RATE
from pyrasgo.utils import polling
from pyrasgo import primitives, schemas, errors
from pyrasgo.schemas.offline import OfflineDataset


class Create:
    """
    API Create Class
    """

    def __init__(self):
        from pyrasgo.api import Get
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self._get = Get()

    def metric(
        self,
        name: str,
        dataset_id: int,
        type: str,
        target_expression: str,
        time_grains: List[Literal["HOUR", "DAY", "WEEK", "MONTH", "QUARTER", "YEAR"]],
        time_dimension: str,
        dimensions: List[str],
        filters: Optional[List[schemas.Filter]] = None,
        meta: Optional[Dict[str, str]] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
    ) -> schemas.Metric:
        """
        Creates and returns a metric on a dataset

        Args:
            name: Name of the new metric
            dataset_id: Rasgo ID for the dataset the metric will be built from
            type: Aggregate function to create the metric, e.g., "average" or "sum"
            target_expression: Column name or expression on which to create the metric
            time_grains: Time level at which to apply the metric, e.g., "DAY" or "WEEK"
            time_dimension: Name of date/time column on which to apply the time_grains
            dimensions: Other dimensional column names used in metric calculation
            filters: Filter expressions to apply to the dataset before calculating metric values
            meta: Metadata about the metric to store as attributes
            label: For tagging and organization purposes, add labels to your metrics
            description: Explanatory information about your new metric
        """
        time_grain_objs = [schemas.metric.TimeGrain[grain] for grain in time_grains]
        metric = schemas.MetricCreate(
            ds_dataset_id=dataset_id,
            name=name,
            type=type,
            target_expression=target_expression,
            time_grains=time_grain_objs,
            time_dimension=time_dimension,
            dimensions=dimensions,
            filters=filters,
            meta=meta,
            label=label,
            description=description,
        )
        try:
            response = self.api._post(
                "/metric", metric.dict(exclude_unset=True, exclude_none=True), api_version=2
            ).json()
            return schemas.Metric(**response)
        except Exception as err:
            raise errors.RasgoResourceException(f"Could not create metric on dataset '{dataset_id}'.") from err

    def transform(
        self,
        *,
        name: str,
        source_code: str,
        type: Optional[str] = None,
        arguments: Optional[List[dict]] = None,
        description: Optional[str] = None,
        tags: Optional[Union[List[str], str]] = None,
        context: Optional[Dict[str, Any]] = None,
        dw_type: Optional[Literal["SNOWFLAKE", "BIGQUERY", "UNSET"]] = None,
    ) -> schemas.Transform:
        """
        Create and return a new Transform in Rasgo
        Args:
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform
            tags: List of tags, or a tag (string), to set on this dataset
            context: Object used to add context to transforms for client use
            dw_type: DataWarehouse provider: SNOWFLAKE, BIGQUERY or UNSET
                     if not provided, will be set to your current DataWarehouse

        Returns:
            Created Transform obj
        """
        arguments = arguments if arguments else []

        # Init tag array to be list of strings
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        transform = schemas.TransformCreate(
            name=name,
            type=type,
            source_code=source_code,
            description=description,
            tags=tags,
            context=context,
            dw_type=dw_type.upper() if dw_type else None,
        )
        transform.arguments = [schemas.TransformArgumentCreate(**x) for x in arguments]
        response = self.api._post("/transform", transform.dict(), api_version=2).json()
        return schemas.Transform(**response)

    # ----------------------------------
    #  Internal/Private Create Calls
    # ----------------------------------
    def _dataset_from_draft(
        self,
        *,
        dataset_contract: schemas.DatasetPublish,
        timeout: Optional[int] = None,
    ) -> schemas.Dataset:
        """
        Calls Rasgo's API to publish a Dataset from a Draft.

        Args:
            dataset_contract: publish contract to send to the API
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
        Returns:
            Dataset object, polled from async API
        """
        resp = self.api._post(
            "/datasets/publish", dataset_contract.dict(exclude_unset=True, exclude_none=True), api_version=2
        ).json()
        status_tracking = schemas.StatusTracking(**resp)

        return polling.poll_dataset_publish(
            connection_obj=self.api,
            max_poll_attempts=MAX_POLL_ATTEMPTS,
            status_tracking_obj=status_tracking,
            timeout=timeout,
            poll_retry_rate=POLL_RETRY_RATE,
        )

    def _dataset_from_offline_schema(
        self,
        dataset_instructions: OfflineDataset,
        verbose: bool = False,
        timeout: int = None,
    ) -> primitives.Dataset:
        """
        Creates a new dataset based on a Rasgo-compliant dict

        Args:
            dataset_dict: a Rasgo-compliant json dict (converted from a yaml file)
        """
        response = self.api._post("/datasets/from-offline-version", dataset_instructions.dict(), api_version=2).json()
        if verbose:
            print(f"Request sent to republish Dataset {dataset_instructions.resource_key}, polling for response...")
        status_tracking = schemas.StatusTracking(**response)
        return polling.poll_dataset_publish(
            connection_obj=self.api,
            max_poll_attempts=MAX_POLL_ATTEMPTS,
            status_tracking_obj=status_tracking,
            timeout=timeout,
            poll_retry_rate=POLL_RETRY_RATE,
        )

    def _dataset_from_table(
        self,
        *,
        fqtn: str,
        dataset_contract: schemas.DatasetCreate,
        timeout: Optional[int] = None,
    ) -> schemas.Dataset:
        """
        Calls Rasgo's API to publish a Dataset from a Table.

        Args:
            fqtn: fully-qualified table name
            dataset_publish_in: publish contract to send to the API
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
        Returns:
            Dataset object, polled from async API
        """
        resp = self.api._post(
            f"/datasets/async?fqtn={parse.quote(fqtn)}",
            dataset_contract.dict(exclude_unset=True, exclude_none=True),
            api_version=2,
        ).json()
        status_tracking = schemas.StatusTracking(**resp)

        return polling.poll_dataset_publish(
            connection_obj=self.api,
            max_poll_attempts=MAX_POLL_ATTEMPTS,
            status_tracking_obj=status_tracking,
            timeout=timeout,
            poll_retry_rate=POLL_RETRY_RATE,
        )

    def _operation_set_non_async(
        self, operations: List[schemas.OperationCreate], dataset_dependency_ids: List[int]
    ) -> schemas.OperationSet:
        """
        Create an operation set in Rasgo with specified operation
        and input dataset dependencies ids, in a  non-async status
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations, dataset_dependency_ids=dataset_dependency_ids, use_custom_sql=False
        )
        response = self.api._post("/operation-sets", operation_set_create.dict(), api_version=2).json()
        return schemas.OperationSet(**response)

    def _operation_set_async(
        self, operations: List[schemas.OperationCreate], dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetAsyncTask:
        """
        Create an operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations, dataset_dependency_ids=dataset_dependency_ids, use_custom_sql=False
        )
        response = self.api._post("/operation-sets/async", operation_set_create.dict(), api_version=2).json()
        return schemas.OperationSetAsyncTask(**response)

    def _operation_set(
        self,
        operations: List[schemas.OperationCreate],
        dataset_dependency_ids: List[int],
        async_compute: bool = True,
        async_verbose: bool = False,
    ) -> schemas.OperationSet:
        """
        Create and return an Operation set based on the input
        operations and dataset dependencies

        Set param `async_compute` to False to not create op with async

        Args:
            operations: List of operations to add to operation set.
                         Should be in ordered by time operation added.
            dataset_dependency_ids: Dataset ids to set as a parent for this operation set
            async_compute: Set to False not create op set in async fashion in backend/API
            async_verbose: If creating op set in async mode, set verbose to True to have verbose output

        Returns:
            Created Operation Set
        """
        if async_compute:
            # Submit the task request
            task_request = self._operation_set_async(
                operations=operations, dataset_dependency_ids=dataset_dependency_ids
            )
            operation_set_id = polling.poll_operation_set_async_status(task_request=task_request, verbose=async_verbose)
            return self._get._operation_set(operation_set_id)
        else:
            return self._operation_set_non_async(operations=operations, dataset_dependency_ids=dataset_dependency_ids)



"""
API Publish Functions
"""
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal

import numpy as np
import pandas as pd

from pyrasgo import errors
from pyrasgo.primitives.dataset import Dataset
from pyrasgo.schemas.dataset import DatasetCreate, DatasetPublish, DatasetRePublish, DatasetSourceType
from pyrasgo.schemas.offline import OfflineDataset
from pyrasgo.utils import naming


class Publish:
    """
    API Publish Class
    """

    def __init__(self):
        from pyrasgo.api import Create, Get, Update
        from pyrasgo.api.connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()
        self.create = Create()
        self.update = Update()
        self._dw = None

    @property
    def data_warehouse(self):
        from pyrasgo.storage import DataWarehouse, dw_connection

        if self._dw:
            return self._dw
        self._dw: DataWarehouse = dw_connection(self.api._dw_type())
        return self._dw

    def dataset(
        self,
        dataset: Dataset,
        name: str,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        attributes: Optional[dict] = None,
        table_type: Optional[str] = "VIEW",
        table_name: Optional[str] = None,
        if_exists: str = "fail",
        verbose: bool = False,
        timeout: Optional[int] = None,
    ) -> Dataset:
        """
        Saves a transformed Dataset in Rasgo to published
        Args:
            dataset: Dataset to save
            name: Name of dataset
            resource_key: A table-safe key used to identify this dataset
            description: Description of dataset
            attributes: Dictionary with metadata about the Dataset
            table_type: Type of object to create in snowflake. Can be "TABLE" or "VIEW"
            table_name: Data Warehouse Table Name to set for this DS's published operation
            if_exists: Values('fail', 'overwrite') Instructions for how to proceed if an existing Dataset matches this resource_key
                If 'overwrite', Re-Publishes the existing Dataset with this draft
                If 'fail', raises a warning about the key conflict
            verbose: If true will print save progress status
            timeout: Approximate timeout in seconds. Raise an APIError if the dataset isn't available in x seconds
        Returns:
            Dataset
        """
        if verbose:
            print(f"Saving Dataset with name={name!r} description={description!r} resource_key={resource_key}...")

        # Fail if incoming dataset already exists in API
        if dataset._status in ("DRAFT DATASET", "PUBLISHED DATASET"):
            raise errors.RasgoRuleViolation(
                f"This Dataset already exists in Rasgo as resource_key {dataset.resource_key}. "
                "Transform the dataset to save it."
            )

        # Get or create the Operation Set so we can publish it
        elif dataset._status == "API DRAFT":
            dataset._refresh()
            operation_set = dataset._api_operation_set

        elif dataset._status == "OFFLINE DRAFT":
            operation_set = self.create._operation_set(
                operations=dataset._operations,
                dataset_dependency_ids=dataset._dataset_dependencies,
            )

        # Publish vs RePublish routing
        # If this user passes a resource_key check if it exists
        republish = False
        if resource_key:
            try:
                existing_ds = self.get.dataset(resource_key=resource_key)
            except errors.RasgoResourceException:
                existing_ds = None
            if existing_ds:
                if if_exists == "fail":
                    raise errors.DWResourceException(
                        f"A Dataset with resource_key {resource_key} already exists. "
                        "To create a new dataset, re-run with function without the resource_key param or pass in a unique key. "
                        "To overwrite this existing Dataset with your draft, re-run this function with param `if_exists='overwrite'`"
                    )
                elif if_exists == "overwrite":
                    republish = True
                else:
                    raise errors.ParameterValueError("if_exists", ["fail", "overwrite"])

        if republish:
            if verbose:
                print(f"RePublishing Dataset {resource_key}")
            dataset_contract = DatasetRePublish(
                operation_set_resource_key=operation_set.resource_key,
                terminal_operation_resource_key=operation_set.terminal_operation_resource_key,
                name=name,
                description=description,
                attributes=attributes,
                source_type=DatasetSourceType.RASGO,
                table_type=table_type,
            )
            ds = self.update._republish_dataset(
                resource_key=resource_key, dataset_contract=dataset_contract, verbose=verbose, timeout=timeout
            )
        else:
            if verbose:
                print("Publishing new Dataset")
            dataset_obj = DatasetPublish(
                resource_key=resource_key,
                operation_set_resource_key=operation_set.resource_key,
                terminal_operation_resource_key=operation_set.terminal_operation_resource_key,
                table_name=table_name,
                table_type=table_type,
                name=name,
                description=description,
                attributes=attributes,
                source_type=DatasetSourceType.RASGO,
            )
            ds = self.create._dataset_from_draft(dataset_contract=dataset_obj, timeout=timeout)
        if verbose:
            print(f"Dataset {ds.resource_key} saved")
        return Dataset(api_dataset=ds, api_operation_set=operation_set)

    def dataset_from_dict(
        self,
        dataset_dict: dict,
        if_exists: str = "fail",
        verbose: bool = False,
        timeout: int = None,
    ) -> Dataset:
        """
        Publish a new dataset based on a Rasgo-compliant dict

        Args:
            dataset_dict: a Rasgo-compliant json dict (converted from a yaml file)
            if_exists: Values('fail', 'overwrite') Instructions for how to proceed if an existing Dataset matches this resource_key
                If 'overwrite', Re-Publishes the existing Dataset with this draft
                If 'copy', create a new Dataset based on this draft
                If 'fail', raises a warning about the key conflict
            verbose: If true will print save progress status
            timeout: Approximate timeout in seconds. Raise an APIError if the dataset isn't available in x seconds

        To publish a dataset from a yaml file:
            with open(your_file_path, "r") as yaml_file:
                rasgo.publish.dataset_from_dict(yaml.safe_load(yaml_file))
        """
        try:
            dataset_instructions = OfflineDataset(**dataset_dict)
        except Exception as ex:
            raise errors.RasgoSchemaException("dataset_dict violates Rasgo-compliant format") from ex

        # Publish vs RePublish routing
        republish = False
        if dataset_instructions.resource_key:
            try:
                existing_ds = self.get.dataset(resource_key=dataset_instructions.resource_key)
            except errors.RasgoResourceException:
                existing_ds = None
            if existing_ds:
                if if_exists == "fail":
                    raise errors.DWResourceException(
                        f"A Dataset with resource_key {dataset_instructions.resource_key} already exists. "
                        "To create a new dataset, re-run with function without the resource_key param or pass in a unique key. "
                        "To overwrite this existing Dataset with your draft, re-run this function with param `if_exists='overwrite'`"
                    )
                if if_exists == "copy":
                    republish = False
                elif if_exists == "overwrite":
                    republish = True
                else:
                    raise errors.ParameterValueError("if_exists", ["fail", "overwrite", "copy"])

        if republish:
            if verbose:
                print(f"Republishing Dataset {dataset_instructions.resource_key} from this dict")
            return self.update._dataset_from_offline_schema(
                dataset_instructions=dataset_instructions, verbose=verbose, timeout=timeout
            )
        else:
            if verbose:
                print("Creating a new Dataset from this dict")
            return self.create._dataset_from_offline_schema(
                dataset_instructions=dataset_instructions, verbose=verbose, timeout=timeout
            )

    def dbt_project(
        self,
        datasets: List[Dataset],
        project_directory: Union[os.PathLike, str] = None,
        models_directory: Union[os.PathLike, str] = None,
        project_name: str = None,
        model_args: Dict[str, Any] = None,
        verbose: bool = False,
    ) -> str:
        """
        Exports all given datasets to Models in a dbt Project

        Params:
        `datasets`: List[Dataset]:
            list of Rasgo datasets to write to dbt as models
        `project_directory`: Path:
            directory to save project files to
            defaults to current working dir
        `models_directory`: Path:
            directory to save model files to
            defaults to project_directory/models
        `project_name`: str:
            name for this project
            defaults to organization name
        """
        from pyrasgo.primitives import DbtProject
        from pyrasgo.utils.dbt import dataset_to_model, dataset_to_source

        project_directory = Path(project_directory or os.getcwd())
        project_name = project_name or naming.cleanse_dbt_name(self.api._profile["organization"]["name"])
        models_directory = Path(models_directory or (project_directory / "models"))
        project = DbtProject(
            name=project_name,
            project_directory=project_directory,
            models_directory=models_directory,
            models=[dataset_to_model(ds) for ds in datasets if not ds.is_source],
            sources=[dataset_to_source(ds) for ds in datasets if ds.is_source],
            model_args=model_args,
        )
        return project.save_files(verbose=verbose)

    def df(
        self,
        df: pd.DataFrame = None,
        name: Optional[str] = None,
        resource_key: Optional[str] = None,
        fqtn: Optional[str] = None,
        description: Optional[str] = None,
        attributes: Optional[dict] = None,
        category: Optional[str] = None,
        parents: Optional[List[Dataset]] = None,
        verbose: bool = False,
        if_exists: Optional[Literal["append", "overwrite", "fail"]] = "fail",
    ) -> Dataset:
        """
        Push a Pandas Dataframe a Data Warehouse table and register it as a Rasgo Dataset

        params:
            df: pandas DataFrame
            name: Optional name for the Dataset (if not provided a random string will be used)
            description: Optional description for this Rasgo Dataset
            parents: Set Parent Dataset dependencies for this df dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            category: user-defined string to categorize this Dataset
            fqtn: Optional name for the target table (if not provided a random string will be used)
            if_exists: Values: ['fail', 'append', 'overwrite'] directs the function what to do if a FTQN is passed, and represents an existing Dataset
        return:
            Rasgo Dataset
        """
        # Make sure no incompatible dw dtypes in df uploading
        _raise_error_if_bad_df_dtypes(df)

        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if_exists_vals = ["overwrite", "append", "fail"]
        if_exists = if_exists.lower()
        if if_exists not in if_exists_vals:
            raise errors.ParameterValueError("if_exists", if_exists_vals)

        # If FQTN or resource_key are passed, try to match existing datasets
        ds_matched_on_rk = None
        ds_matched_on_fqtn = None
        if resource_key:
            try:
                ds_matched_on_rk = self.get.dataset(resource_key=resource_key)
            except errors.RasgoResourceException:
                ds_matched_on_rk = None
        if fqtn:
            naming.is_fqtn(fqtn, raise_if_false=True)
            try:
                ds_matched_on_fqtn = self.get.dataset(fqtn=fqtn)
            except errors.RasgoResourceException:
                ds_matched_on_fqtn = None

        # If a match is found, make sure the FQTN + rk combo does not conflict with Dataset on record
        if ds_matched_on_rk and ds_matched_on_fqtn:
            if ds_matched_on_rk.id != ds_matched_on_fqtn.id:
                raise errors.ParameterValueError(
                    message="`resource_key` and `fqtn` args were both passed, but do not match. "
                    f"resource_key {resource_key} references Dataset {ds_matched_on_rk.id} & "
                    f"fqtn {fqtn} references Dataset {ds_matched_on_fqtn.id} "
                    "To update the Dataset referenced by this resource_key, please omit or change your fqtn arg. "
                    "To update the Dataset referenced by this fqtn, please omit or change your resource_key arg."
                )
        if ds_matched_on_rk and fqtn:
            if ds_matched_on_rk.fqtn != fqtn:
                raise errors.ParameterValueError(
                    message="`resource_key` and `fqtn` args were both passed, but do not match. "
                    f"resource_key {resource_key} references Dataset {ds_matched_on_rk.id}. "
                    f"Its fqtn is {ds_matched_on_rk.fqtn}, which does not match the arg {fqtn}. "
                    "To update the Dataset referenced by this resource_key, please omit or change your fqtn arg. "
                    "To update the Dataset referenced by this fqtn, please omit or change your resource_key arg."
                )
        if ds_matched_on_fqtn and resource_key:
            if ds_matched_on_fqtn.resource_key != resource_key:
                raise errors.ParameterValueError(
                    message="`resource_key` and `fqtn` args were both passed, but do not match. "
                    f"fqtn {fqtn} references Dataset {ds_matched_on_fqtn.id}. "
                    f"Its resource_key is {ds_matched_on_fqtn.resource_key}, which does not match the arg {resource_key}. "
                    "To update the Dataset referenced by this fqtn, please omit or change your resource_key arg. "
                    "To update the Dataset referenced by this resource_key, please omit or change your fqtn arg."
                )

        ds_to_update = ds_matched_on_rk or ds_matched_on_fqtn
        if ds_to_update:
            # Users should only be able to overwrite datasets in their own organization
            if ds_to_update._api_dataset.organization_id != self.api._profile.get("organizationId"):
                raise errors.RasgoRuleViolation(
                    f"Dataset {ds_to_update.id} already exists. This API key does not have permission to replace it."
                )
            if if_exists == "fail":
                raise errors.ParameterValueError(
                    message=f"Dataset {ds_to_update.id} already exists, and {if_exists} was passed for `if_exists`. "
                    "Please confirm the resource_key & fqtn or choose another value for `if_exists`"
                )
            print(f"Found Dataset {ds_to_update.id} matching your input args. Proceeding in {if_exists} mode.")

        # Determine the table name this df will be written into
        if not resource_key:
            resource_key = naming.random_alias()
        if not fqtn:
            target_database, target_schema, target_table = naming.split_fqtn(
                resource_key, self.api._default_dw_namespace()
            )
            fqtn = naming.make_fqtn(database=target_database, schema=target_schema, table=target_table)

        # Write the df to the target table
        if verbose:
            verb = "Appending" if if_exists == "append" else "Writing"
            print(f"{verb} dataframe to target table {fqtn}")
        self.data_warehouse.write_dataframe_to_table(df, table_name=fqtn, method=if_exists)

        # Return the dataset if it already exists
        if ds_to_update:
            return ds_to_update

        # Otherwise, create a dataset based on the new table created from df
        return self.table(
            fqtn=fqtn,
            name=name or fqtn,
            resource_key=resource_key,
            description=description,
            verbose=verbose,
            attributes=attributes,
            category=category,
            parents=parents,
            source_type=DatasetSourceType.DATAFRAME,
        )

    def table(
        self,
        fqtn: str,
        name: Optional[str] = None,
        resource_key: Optional[str] = None,
        description: Optional[str] = None,
        parents: Optional[List[Dataset]] = None,
        verbose: bool = False,
        attributes: Optional[dict] = None,
        category: Optional[str] = None,
        if_exists: Optional[Literal["return", "fail"]] = "fail",
        timeout: Optional[int] = None,
        source_type: DatasetSourceType = DatasetSourceType.TABLE,
    ) -> Dataset:
        """
        Register an existing table as a Rasgo Dataset

        params:
            fqtn: The fully qualified table name of the table to register
            name: Optional name to apply to this Rasgo Dataset
            description: Optional description for this Rasgo Dataset
            parents: Set Parent Dataset dependencies for this table dataset. Input as list of dataset primitive objs.
            verbose: Print status statements to stdout while function executes if true
            attributes: Dictionary with metadata about the Dataset
            category: user-defined string to categorize this Dataset
            if_exists: Values: ['fail', 'return'] directs the function what to do if a FTQN is passed, and represents an existing Dataset
            timeout: Approximate timeout for creating the table in seconds. Raise an APIError if the reached
            source_type: Specifies the `source_type` of this Dataset in Rasgo. Defaults to 'TABLE'
        return:
            Rasgo Dataset
        """
        # Validate all parent ds Ids exist if passed
        # Calling rasgo.get.dataset(<id>) will raise error if doesn't
        parents = parents if parents else []
        parent_ids = [ds.id for ds in parents]
        for p_ds_id in parent_ids:
            self.get.dataset(p_ds_id)

        if verbose:
            print(f"Publishing {source_type.value} {fqtn} as Rasgo dataset")

        naming.is_fqtn(fqtn, raise_if_false=True)

        table_database, table_schema, table_name = naming.split_fqtn(fqtn)

        try:
            row_count = self.data_warehouse.query_into_dict(
                f"select count(1) as ROW_CT from {table_database}.{table_schema}.{table_name}"
            )
            if row_count[0]["ROW_CT"] == 0:
                raise errors.DWResourceException(
                    f"Source table {table_name} is empty or this role does not have access to it."
                )
        except Exception as err:
            raise errors.DWResourceException(
                f"Source table {table_name} does not exist or this role does not have access to it."
            ) from err

        # Make sure `source_type` param is valid Enum Value
        if not isinstance(source_type, DatasetSourceType):
            raise errors.ParameterValueError("source_type", ["CSV", "RASGO", "TABLE", "DATAFRAME"])

        # Check if a Dataset already exists
        existing_ds = None
        try:
            existing_ds = self.get.dataset(fqtn=fqtn)
        except errors.RasgoResourceException:
            pass
        if existing_ds:
            if if_exists == "return":
                return existing_ds
            else:  # if_exists == "fail":
                raise errors.RasgoResourceException(
                    f"{fqtn} is already registered with Rasgo as Dataset {existing_ds.id}"
                )

        # Create operation set with parent dependencies
        # set for this dataset
        operation_set = self.create._operation_set(
            operations=[],
            dataset_dependency_ids=parent_ids,
            async_compute=False,
        )

        # Publish Dataset with operation set created above
        dataset = self.create._dataset_from_table(
            fqtn=fqtn,
            dataset_contract=DatasetCreate(
                name=name or table_name,
                resource_key=resource_key,
                dw_operation_set=operation_set.id,
                source_type=source_type.value,
                description=description,
                attributes=attributes,
                category=category,
            ),
            timeout=timeout,
        )
        # Raise API error if backend error creating dataset
        if not dataset:
            raise errors.APIError("DataSource failed to upload")

        # Return dataset if no error
        return Dataset(api_dataset=dataset)


# ------------------------------------------
#  Private Helper Funcs for Publish Class
# ------------------------------------------


def _raise_error_if_bad_df_dtypes(df: pd.DataFrame) -> None:
    """
    Raise an API error is any dtypes in the pandas dataframe,
    which are being pushed to the data warehouse aren't compatible.

    Raise proper error message if so
    """
    for col_name in df:
        col = df[col_name]
        if col.dtype.type == np.datetime64:
            raise errors.RasgoRuleViolation(
                "Can't publish pandas Df to Rasgo. Df column "
                f"'{col_name}' needs to be converted to proper datetime format.\n\n"
                "If your column is a **DATE** use `pd.to_datetime(df[<col_name>]).dt.date` to convert it\n"
                "If your column is a **TIMESTAMP** use `pd.to_datetime(final_df['col_name']).dt.tz_localize('UTC')`"
            )



from .attributes import Attribute
from .dataset import (
    Dataset,
    DatasetBulk,
    DatasetCreate,
    DatasetUpdate,
    DatasetSourceType,
    DatasetPublish,
    DatasetRePublish,
)
from .dataset_column import DatasetColumn, DatasetColumnUpdate
from .dw_operation import Operation, OperationCreate
from .dw_operation_set import (
    OperationSet,
    OperationSetAsyncEvent,
    OperationSetAsyncTask,
    OperationSetCreate,
    BulkOperationSet,
)
from .dw_table import DataColumn, DataTable, DataTableWithColumns
from .enums import OperationSetAsyncTaskType
from .jobs import AsyncUpdateTask
from .organization import Organization
from .status_tracking import StatusTracking
from .transform import (
    Transform,
    TransformArgument,
    TransformArgumentCreate,
    TransformCreate,
    TransformExecute,
    TransformUpdate,
)
from .user import User
from .metric import Metric, MetricCreate, MetricUpdate, TimeGrain, Filter
from .offline import OfflineDataset
